from __future__ import annotations

import random

import pytest
from zeus.optimizer.batch_size.common import (
    DELETE_JOB_URL,
    GET_NEXT_BATCH_SIZE_URL,
    REGISTER_JOB_URL,
    REPORT_END_URL,
    REPORT_RESULT_URL,
)

# https://fastapi.tiangolo.com/tutorial/testing/


def test_register_job(client, helpers):
    job_config = helpers.get_fake_job_config("test_register_job")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    print(response.text)
    print(str(response))
    assert response.status_code == 201

    response = client.post(REGISTER_JOB_URL, json=job_config)
    print(response.text)
    assert response.status_code == 200

    job_config["default_batch_size"] = 512
    response = client.post(REGISTER_JOB_URL, json=job_config)
    print(response.text)
    assert response.status_code == 409


def test_register_job_validation_error(client, helpers):
    temp = helpers.get_fake_job_config("test_register_job_validation_error")
    temp["default_batch_size"] = 128
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422

    temp["default_batch_size"] = 0
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422

    temp = helpers.get_fake_job_config("test_register_job_validation_error")
    temp["max_epochs"] = 0
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422

    temp = helpers.get_fake_job_config("test_register_job_validation_error")
    temp["batch_sizes"] = []
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422

    temp = helpers.get_fake_job_config("test_register_job_validation_error")
    temp["eta_knob"] = 1.1
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422

    temp = helpers.get_fake_job_config("test_register_job_validation_error")
    temp["beta_knob"] = 0
    response = client.post(REGISTER_JOB_URL, json=temp)
    assert response.status_code == 422


def test_predict(client, helpers):
    job_config = helpers.get_fake_job_config("test_predict")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    cur_default_bs = job_config["default_batch_size"]
    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    print(response.text)
    assert response.status_code == 200
    assert response.json()["batch_size"] == cur_default_bs
    assert response.json()["trial_number"] == 1

    # concurrent job submission
    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    print(response.text)
    assert response.status_code == 200
    assert response.json()["batch_size"] == cur_default_bs
    assert response.json()["trial_number"] == 2


def test_report(client, helpers):
    job_config = helpers.get_fake_job_config("test_report")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    # Converged within max epoch => successful training
    response = client.post(
        REPORT_RESULT_URL,
        json={
            "job_id": job_config["job_id"],
            "batch_size": response.json()["batch_size"],
            "trial_number": response.json()["trial_number"],
            "time": 14.438,
            "energy": 3000.123,
            "metric": 0.55,
            "current_epoch": 98,
        },
    )
    assert (
        response.status_code == 200
        and response.json()["converged"] == True
        and response.json()["stop_train"] == True
    )

    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    # Early stop
    response = client.post(
        REPORT_RESULT_URL,
        json={
            "job_id": job_config["job_id"],
            "batch_size": response.json()["batch_size"],
            "trial_number": response.json()["trial_number"],
            "time": 30,
            "energy": 6000,
            "metric": 0.55,
            "current_epoch": 98,
        },
    )
    assert (
        response.status_code == 200
        and response.json()["converged"] == False
        and response.json()["stop_train"] == True
    ), response.text


def test_exploration_stage(client, helpers):
    job_config = helpers.get_fake_job_config("test_exploration_stage")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    cur_default_bs = job_config["default_batch_size"]
    bss = job_config["batch_sizes"]
    trial_number = 1
    for trial in range(1, job_config["num_pruning_rounds"] + 1):
        idx = bss.index(cur_default_bs)
        down = sorted(bss[: idx + 1], reverse=True)
        up = sorted(bss[idx + 1 :])
        new_bss = []

        print("Exploration space:", [down, up])
        for bs_list in [down, up]:
            for bs in bs_list:
                # Predict
                response = client.get(
                    GET_NEXT_BATCH_SIZE_URL,
                    params={"job_id": job_config["job_id"]},
                )
                assert response.status_code == 200
                assert response.json()["batch_size"] == bs
                assert response.json()["trial_number"] == trial_number
                trial_number += 1

                # Concurrent job
                response = client.get(
                    GET_NEXT_BATCH_SIZE_URL,
                    params={"job_id": job_config["job_id"]},
                )
                assert response.status_code == 200
                assert (
                    response.json()["batch_size"] == cur_default_bs
                    if trial == 1 and bs == 512
                    else 512
                )
                assert response.json()["trial_number"] == trial_number
                trial_number += 1

                time = 14.438
                converged = random.choice([True, True, False])
                if bs == 1024:  # make default bs converged for 512
                    converged = True
                if (
                    bs == 512
                ):  # make 512 as the best bs so that we can change the default bs to 512 next round
                    converged = True
                    time = 12
                if converged:
                    new_bss.append(bs)

                response = client.post(
                    REPORT_RESULT_URL,
                    json={
                        "job_id": job_config["job_id"],
                        "batch_size": bs,
                        "error": False,
                        # Report to first trial in the loop
                        "trial_number": trial_number - 2,
                        "time": time,
                        "energy": 3000.123,
                        "metric": 0.55 if converged else 0.33,
                        "current_epoch": 98 if converged else 100,
                    },
                )
                assert (
                    response.status_code == 200
                    and response.json()["converged"] == converged
                    and response.json()["stop_train"] == True
                )
                if not converged:
                    break
        bss = sorted(new_bss)
        cur_default_bs = 512


def test_mab_stage(client, helpers):
    job_config = helpers.get_fake_job_config("test_mab_stage")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    bs_seq = []
    # Previous default batch size is converged
    for _ in range(50):
        # Predict
        response = client.get(
            GET_NEXT_BATCH_SIZE_URL,
            params={"job_id": job_config["job_id"]},
        )
        assert response.status_code == 200
        bs = response.json()["batch_size"]
        trial_number = response.json()["trial_number"]
        bs_seq.append(response.json()["batch_size"])
        # Concurrent job
        response = client.get(
            GET_NEXT_BATCH_SIZE_URL,
            params={"job_id": job_config["job_id"]},
        )
        assert response.status_code == 200
        bs_seq.append(response.json()["batch_size"])

        response = client.post(
            REPORT_RESULT_URL,
            json={
                "job_id": job_config["job_id"],
                "batch_size": bs,
                "trial_number": trial_number,
                "error": False,
                "time": 15.123,
                "energy": 3000.123,
                "max_power": 300,
                "metric": 0.55,
                "current_epoch": 98,
            },
        )
        assert (
            response.status_code == 200
            and response.json()["converged"] == True
            and response.json()["stop_train"] == True
        )
    print(bs_seq)


def test_end_trial(client, helpers):
    job_config = helpers.get_fake_job_config("test_end_trial")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    # Start trial
    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    assert response.status_code == 200
    trial_number = response.json()["trial_number"]
    bs = response.json()["batch_size"]

    # End Trial.
    response = client.patch(
        REPORT_END_URL,
        json={
            "job_id": job_config["job_id"],
            "batch_size": bs,
            "trial_number": trial_number,
        },
    )
    assert response.status_code == 200

    # Start trial
    response = client.get(
        GET_NEXT_BATCH_SIZE_URL,
        params={"job_id": job_config["job_id"]},
    )
    assert response.status_code == 200
    trial_number = response.json()["trial_number"]
    bs = response.json()["batch_size"]

    # Report result.
    response = client.post(
        REPORT_RESULT_URL,
        json={
            "job_id": job_config["job_id"],
            "batch_size": bs,
            "trial_number": trial_number,
            "error": False,
            "time": 15.123,
            "energy": 3000.123,
            "max_power": 300,
            "metric": 0.55,
            "current_epoch": 98,
        },
    )

    # End Trial.
    response = client.patch(
        REPORT_END_URL,
        json={
            "job_id": job_config["job_id"],
            "batch_size": bs,
            "trial_number": trial_number,
        },
    )
    assert response.status_code == 200


def test_delete_job(client, helpers):
    job_config = helpers.get_fake_job_config("test_delete_job")
    response = client.post(REGISTER_JOB_URL, json=job_config)
    assert response.status_code == 201

    response = client.delete(
        DELETE_JOB_URL,
        params={"job_id": job_config["job_id"]},
    )
    print(response.text)
    assert response.status_code == 200

    response = client.post(REGISTER_JOB_URL, json=job_config)
    print(response.text)
    assert response.status_code == 201

    response = client.delete(
        DELETE_JOB_URL,
        params={"job_id": job_config["job_id"]},
    )
    print(response.text)
    assert response.status_code == 200

    # Job doesn't exists
    response = client.delete(
        DELETE_JOB_URL,
        params={"job_id": "UNKNOWN"},
    )
    print(response.text)
    assert response.status_code == 404



"""Example script of running pipeline frequency optimization."""

from __future__ import annotations

import time
import itertools
import logging
from pathlib import Path
from typing import Type
from collections import defaultdict
from dataclasses import dataclass

import tyro
import pandas as pd
import networkx as nx
import matplotlib.pyplot as plt

from lowtime.operation import (
    CandidateExecutionOptions,
    OperationSpec,
    ExecutionOption,
)
from lowtime.cost_model import ExponentialModel
from lowtime.perseus.instruction import (
    Instruction,
    Forward,
    Backward,
    forward_dep,
    backward_dep,
)
from lowtime.solver import PhillipsDessouky
from lowtime.graph_utils import add_source_node, add_sink_node, DependencyResolver
from lowtime.perseus.schedule import Synchronous1F1B
from lowtime.perseus.visualizer import PipelineVisualizer, ANNOTATE_ARGS, LINE_ARGS

logger = logging.getLogger()


@dataclass
class Args:
    # Path to instruction profile results
    inst_profile: str
    # Directory to output results
    output_dir: Path
    # Number of microbatchs
    num_mbs: int
    # Number of stages
    num_stages: int
    # GPU power consumption while blocking on P2P communication, in Watts
    p2p_power: float = 70.0
    # Interval to draw the state of the pipeline
    plot_interval: int = 100
    # The unit of reduction for each iteration, in seconds
    unit_time: float = 0.001
    # Noise factor for soft Pareto frontier filtering
    noise_factor: float = 0.95


def main(args: Args) -> None:
    """Perseus time-cost tradeoff optimization."""
    # Setup logging and output.
    output_dir = Path(args.output_dir)
    output_dir.mkdir(parents=True, exist_ok=False)
    log_path = output_dir / "job.log"

    logging.basicConfig(
        format="%(asctime)s %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
        level=logging.INFO,
        handlers=[logging.FileHandler(log_path, mode="a"), logging.StreamHandler()],
    )
    logger.info("Arguments: %s", args)

    # Instruction offline profiling results.
    inst_df = pd.read_csv(args.inst_profile)

    ####################
    # Execution Option #
    ####################
    # Construct the OperationSpec object of each pipeline instruction in each stage.
    op_spec_map: dict[int, dict[Type[Instruction], OperationSpec]] = defaultdict(dict)
    for instruction in [Forward, Backward]:
        inst_name = instruction.__name__
        for stage_id in range(args.num_stages):
            logger.info("Processing %s stage %d", inst_name, stage_id)
            options = []
            _df = inst_df.query(
                f"stage == {stage_id} and instruction == '{inst_name.lower()}'"
            )
            for _, row in _df.iterrows():
                row = row.to_dict()
                options.append(
                    ExecutionOption[int](
                        real_time=row["time"],
                        unit_time=args.unit_time,
                        cost=row["energy"],
                        knob=int(row["frequency"]),
                    )
                )

            # Map the cost to be effective computation energy.
            # Everything from now on is in terms of effective energy.
            for option in options:
                option.cost -= args.p2p_power * option.quant_time * option.unit_time

            # Get the Preto frontier, quantize time, and deduplicate time.
            cand_options = CandidateExecutionOptions[int](options=options)
            if len(cand_options.options) <= 3:
                cand_options = CandidateExecutionOptions[int](
                    options=options, noise_factor=args.noise_factor
                )

            # Fit the cost model.
            model = ExponentialModel(cand_options, search_strategy="best")

            # Draw the cost model.
            fig, ax = plt.subplots(figsize=(8, 8), tight_layout=True)
            model.draw(ax, cand_options)
            fig.savefig(f"{output_dir}/{inst_name.lower()}_{stage_id}.png")

            # Initialize the operation spec.
            op_spec = OperationSpec[int](options=cand_options, cost_model=model)
            op_spec_map[stage_id][instruction] = op_spec

    ####################
    # DAG construction #
    ####################
    dag = nx.DiGraph()

    # Generate and add all instructions to the DAG.
    # Reserve 0 for dummy source and 1 for dummy sink.
    node_id = 2
    instructions: list[list[Instruction]] = []
    for stage_id in range(args.num_stages):
        # Generate instructions for each stage.
        stage_insts: list[Instruction] = []
        stage_node_ids: list[int] = []
        for inst in Synchronous1F1B(
            num_stages=args.num_stages,
            num_micro_batches=args.num_mbs,
            stage_id=stage_id,
            operation_spec_map=op_spec_map[stage_id],
        ):
            dag.add_node(node_id, op=inst)
            stage_insts.append(inst)
            stage_node_ids.append(node_id)
            node_id += 1
        instructions.append(stage_insts)

        # Add dependencies between adjacent instructions in the same stage.
        for node_id1, node_id2 in zip(stage_node_ids, stage_node_ids[1:]):
            dag.add_edge(node_id1, node_id2)

    # Add dependencies between dependent pipeline instructions.
    insts = dag.nodes(data=True)
    resolver = DependencyResolver(
        dependency_rules=[forward_dep, backward_dep],
        node_type=Instruction,
    )
    for (id1, data1), (id2, data2) in itertools.product(insts, insts):
        if resolver.is_dependent(data1["op"], data2["op"]):
            dag.add_edge(id1, id2)

    # Add source and sink nodes.
    add_source_node(dag, 0)
    add_sink_node(dag, 1)
    dag.graph["source_node"] = 0
    dag.graph["sink_node"] = 1

    ###################################
    # Time-cost tradeoff optimization #
    ###################################
    def annotation_hook(inst: Instruction) -> str:
        return f"{type(inst).__name__[0]}\n{inst.micro_batch_id}"

    def draw(dag: nx.DiGraph, iteration: int, xlim: int) -> None:
        ANNOTATE_ARGS[Forward]["fontsize"] = 11.0
        ANNOTATE_ARGS[Backward]["fontsize"] = 11.0
        ANNOTATE_ARGS[Forward]["color"] = "black"
        ANNOTATE_ARGS[Backward]["color"] = "black"
        LINE_ARGS["linewidth"] = 3.0

        fig, ax = plt.subplots(figsize=(args.num_mbs, 4), tight_layout=True)

        vis = PipelineVisualizer(dag)
        vis.draw(
            ax,
            draw_time_axis=True,
            p2p_power=args.p2p_power,
            annotation_hook=annotation_hook,
            power_color="RdBu_r",
            normalizer_range=(-200, 550),
        )
        vis.draw_critical_path(ax)

        # Fix xlim so that we can visually see the pipeline width shrink.
        ax.set_xlim(0.0, xlim)
        ax.set_title(f"Iteration {iteration:4d}")
        fig.savefig(f"{output_dir}/pipeline_{iteration:05d}.png")
        plt.close(fig)

    solver = PhillipsDessouky(dag)

    draw_xlim = None
    iteration = 0
    for iteration, result in enumerate(solver.run()):
        # Maybe draw the pipeline.
        if iteration % args.plot_interval == 0:
            if draw_xlim is None:
                draw_xlim = int(result.real_time) + 1
            draw(dag, iteration, draw_xlim)

        # Write the frequency assignment Python file.
        f = open(args.output_dir / f"freqs_pipeline_{iteration:05d}.py", "w")
        f.write("[\n")
        for stage_id, stage_insts in enumerate(instructions):
            stage_freq: list[tuple[str, int]] = []
            for inst in stage_insts:
                stage_freq.append((type(inst).__name__.lower(), inst.assigned_knob))
            f.write(f"{stage_freq},\n")
        f.write("]\n")

        iter_str = f"# Iteration {iteration} "
        real_cost = result.cost + args.num_stages * result.real_time * args.p2p_power
        f.write(iter_str + f"cost change: {result.cost_change}\n")
        f.write(iter_str + f"total cost: {result.cost}\n")
        f.write(iter_str + f"total cost with P2P: {real_cost}\n")

    assert draw_xlim is not None
    draw(dag, iteration, draw_xlim)


if __name__ == "__main__":
    args = tyro.cli(Args)

    start_time = time.time()
    main(args)
    logger.info("Total time: %.2fs", time.time() - start_time)



"""Function for getting latitude and longitude."""

import requests
import logging

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)


def get_ip_lat_long() -> tuple[float, float]:
    """Retrieve the latitude and longitude of the current IP position."""
    try:
        ip_url = "http://ipinfo.io/json"
        resp = requests.get(ip_url)
        loc = resp.json()["loc"]
        lat, long = map(float, loc.split(","))
        logger.info("Retrieved latitude and longitude: %s, %s", lat, long)
        return lat, long
    except requests.exceptions.RequestException as e:
        logger.exception(
            "Failed to retrieve current latitude and longitude of IP: %s", e
        )
        raise
